/*
 * linux/drivers/char/daq8dev.c
 *
 * Driver for DAQ 8ADC system                    
 *
 * Based on the qpmouse driver by Johan Myreen.
 *
 * D.E.Carter 24Jan2000 carter@ohiou.edu
 */

#include <linux/config.h>

#include <linux/module.h>
 
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/fcntl.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/malloc.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/poll.h>
#include <linux/init.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/semaphore.h>

// #include <linux/pc_keyb.h>		/* device enable command.. */


/*
 * We use minor number 1 for DAQ                                      
 */
#define DAQ_MINOR      1	       		/* Minor device # for DAQ */
#define DAQ_BUF_SIZE   32768                    /* was 4096 */

struct daq_queue {
	unsigned long head;
	unsigned long tail;
	struct wait_queue *proc_list;
	struct fasync_struct *fasync;
	unsigned char buf[DAQ_BUF_SIZE];
};

static struct daq_queue *queue;

static unsigned int get_from_queue(void)
{
	unsigned int result;
	unsigned long flags;

	save_flags(flags);
	cli();
	result = queue->buf[queue->tail];
	queue->tail = (queue->tail + 1) & (DAQ_BUF_SIZE-1);
	restore_flags(flags);
	return result;
}


static inline int queue_empty(void)
{
	return queue->head == queue->tail;
}

static int fasync_daq(int fd, struct file *filp, int on)
{
	int retval;

	retval = fasync_helper(fd, filp, on, &queue->fasync);
	if (retval < 0)
		return retval;
	return 0;
}

/*
 *	DAQ Interface
 */

#define DAQ_DATA         0x300		/* Data Port I/O Address */
#define DAQ_STATUS       0x300		/* Status Port I/O Address */

#define DAQ_DEV_IDLE     0x01		/* Device Idle */
#define DAQ_RX_FULL      0x02		/* Device Char received */
#define DAQ_TX_IDLE      0x04		/* Device XMIT Idle */
#define DAQ_RESET        0x08		/* Device Reset */
#define DAQ_INTS_ON      0x10		/* Device Interrupt On */
#define DAQ_ERROR_FLAG   0x20		/* Device Error */
#define DAQ_CLEAR        0x40		/* Device Clear */
#define DAQ_ENABLE       0x80		/* Device Enable */

#define DAQ_IRQ          5

static int daq_present = 0;
static int daq_count = 0;
static int daq_data = DAQ_DATA;
// static int daq_status = DAQ_STATUS;

static int poll_daq_status(void);

/*
 * Interrupt handler for the DAQ port.  Data
 * is waiting in the DAQ FIFO.
 */

static void daq_interrupt(int cpl, void *dev_id, struct pt_regs * regs)
{
        unsigned int data;
        unsigned char ack;
	int head = queue->head;
        int maxhead;
        //printk("Entering interrupt\n");

        do
            {
            maxhead = (queue->tail-2) & (DAQ_BUF_SIZE-1);
            data=inw(0x300);
            if((data & 0x8000) == 0) 
              {
              data=inw(0x300); 
              ack=inb(0x302);
              }
            if ((head != maxhead) && ((head+1) != maxhead))
              {
              queue->buf[head] = (unsigned char)(data & 0x00ff);
              head++;
              head &= DAQ_BUF_SIZE-1;
              //queue->head = head;
              queue->buf[head] = (unsigned char)((data >> 8) & 0x00ff);
              head++;
              head &= DAQ_BUF_SIZE-1;
              //queue->head = head;
              } 
            }while((data & 0x8000) == 0);
        queue->head = head;
	if (queue->fasync)
	kill_fasync(queue->fasync, SIGIO, POLL_IN);
	wake_up_interruptible(&queue->proc_list);
        //printk("Exiting interrupt\n");
}

static int release_daq(struct inode * inode, struct file * file)
{
//	unsigned char status;

	fasync_daq(-1, file, 0);
	if (!--daq_count) {
		if (!poll_daq_status())
			printk("Warning: DAQ device busy in release_daq()\n");
//		status = inb_p(daq_status);
//		outb_p(status & ~(DAQ_ENABLE|DAQ_INTS_ON), daq_status);
//		if (!poll_daq_status())
//			printk("Warning: DAQ device busy in release_daq()\n");
		free_irq(DAQ_IRQ, NULL);
		MOD_DEC_USE_COUNT;
	}
	return 0;
}

/*
 * Install interrupt handler.
 * Enable the device, enable interrupts. 
 */

static int open_daq(struct inode * inode, struct file * file)
{
	unsigned char status;

	if (!daq_present)
		return -EINVAL;

	if (daq_count++)
		return 0;

	if (request_irq(DAQ_IRQ, daq_interrupt, 0, "DAQ Device", NULL)) {
		daq_count--;
		return -EBUSY;
	}

//	status = inb_p(daq_status);
//	status |= (DAQ_ENABLE|DAQ_RESET);
//	outb_p(status, daq_status);
//	status &= ~(DAQ_RESET);
//	outb_p(status, daq_status);

	queue->head = queue->tail = 0;          /* Flush input queue */
	status |= DAQ_INTS_ON;
//	outb_p(status, daq_status);              /* Enable interrupts */

//	while (!poll_daq_status()) {
//		printk("Error: DAQ device busy in open_daq()\n");
//		daq_count--;
//		status &= ~(DAQ_ENABLE|DAQ_INTS_ON);
//		outb_p(status, daq_status);
//		free_irq(DAQ_IRQ, NULL);
//		return -EBUSY;
//	}

//	outb_p(AUX_ENABLE_DEV, daq_data);	/* Wake up device */
	MOD_INC_USE_COUNT;
	return 0;
}

/*
 * Write to the DAQ device.
 */

static ssize_t write_daq(struct file * file, const char * buffer,
			size_t count, loff_t *ppos)
{
	ssize_t i = count;

	while (i--) {
		char c;
		if (!poll_daq_status())
			return -EIO;
		get_user(c, buffer++);
        //      outb_p(c, daq_data);
	}
	file->f_dentry->d_inode->i_mtime = CURRENT_TIME;
	return count;
}

static unsigned int poll_daq(struct file *file, poll_table * wait)
{
	poll_wait(file, &queue->proc_list, wait);
	if (!queue_empty())
		return POLLIN | POLLRDNORM;
	return 0;
}

/*
 * Wait for device to send output char and flush any input char.
 */

#define MAX_RETRIES (60)

static int poll_daq_status(void)
{
/*
	int retries=0;

	while ((inb(daq_status)&(DAQ_RX_FULL|DAQ_TX_IDLE|DAQ_DEV_IDLE))
		       != (DAQ_DEV_IDLE|DAQ_TX_IDLE)
		       && retries < MAX_RETRIES) {

		if (inb_p(daq_status)&(DAQ_RX_FULL))
			inb_p(daq_data);
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout((5*HZ + 99) / 100);
		retries++;
	}
	return !(retries==MAX_RETRIES);
*/
	return (1);
}

/*
 * Put bytes from input queue to buffer.
 */

static ssize_t read_daq(struct file * file, char * buffer,
			size_t count, loff_t *ppos)
{
	struct wait_queue wait = { current, NULL };
	ssize_t i = count;
	unsigned char c;

	if (queue_empty()) {
		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;
		add_wait_queue(&queue->proc_list, &wait);
repeat:
		current->state = TASK_INTERRUPTIBLE;
		if (queue_empty() && !signal_pending(current)) {
			schedule();
			goto repeat;
		}
		current->state = TASK_RUNNING;
		remove_wait_queue(&queue->proc_list, &wait);
	}
	while (i > 0 && !queue_empty()) {
		c = get_from_queue();
		put_user(c, buffer++);
		i--;
	}
	if (count-i) {
		file->f_dentry->d_inode->i_atime = CURRENT_TIME;
		return count-i;
	}
	if (signal_pending(current))
		return -ERESTARTSYS;
	return 0;
}

struct file_operations daq_fops = {
	NULL,		/* seek */
	read_daq,
	write_daq,
	NULL, 		/* readdir */
	poll_daq,
	NULL, 		/* ioctl */
	NULL,		/* mmap */
	open_daq,
	NULL,		/* flush */
	release_daq,
	NULL,
	fasync_daq,
};

/*
 * Initialize driver.
 */
static struct miscdevice daq_device = {
	DAQ_MINOR, "DAQdevice", &daq_fops
};

/*
 * Function to read register in 82C710.
 */

static inline unsigned char read_710(unsigned char index)
{
	outb_p(index, 0x390);			/* Write index */
	return inb_p(0x391);			/* Read the data */
}

int __init daqdevice_init(void)
{
	printk(KERN_INFO "DAQ device detected -- driver installed.\n");
  	printk("DAQ address = %x (should be 0x300)\n", daq_data);   
	daq_present = 1;
	//misc_register(&daq_device);
	 register_chrdev(129,"daq8dev",&daq_fops);
	queue = (struct daq_queue *) kmalloc(sizeof(*queue), GFP_KERNEL);
	memset(queue, 0, sizeof(*queue));
	queue->head = queue->tail = 0;
	queue->proc_list = NULL;

	return 0;
}

int init_module(void)
{
	return daqdevice_init();
}

void cleanup_module(void)
{
	misc_deregister(&daq_device);
	kfree(queue);
}
